/*
 This is an implementation of a subset of the public domain FFT routines of
 Lynn Ten Eyck [Acta Cryst. (1977), A33, 486.], based on an f2c translation
 of the original Fortran routines.
 */

#ifndef FFT_HPP
#define FFT_HPP

#include "geometry.hpp"

#include <cmath>
#include <stdexcept>

void symmetrized_reordering_factoring(int pts, int pmax, int twogrp,
                                      int* factor, int* sym, int* psym,
                                      int* unsym);

template <typename DataType>
void double_inplace_reordering(int pts, int* sym, int* psym, int* unsym,
                               int* dim, DataType* x, DataType* y);

template <typename DataType>
void complex_ft_kernel_driver(int n, int* factor, int* dim, DataType* x,
                              DataType* y);

template <typename DataType>
void radix2_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, int* dim);

template <typename DataType>
void radix3_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, int* dim);

template <typename DataType>
void radix4_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, DataType* x3,
                      DataType* y3, int* dim);

template <typename DataType>
void radix5_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, DataType* x3,
                      DataType* y3, DataType* x4, DataType* y4, int* dim);

template <typename DataType>
void radix8_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, DataType* x3,
                      DataType* y3, DataType* x4, DataType* y4, DataType* x5,
                      DataType* y5, DataType* x6, DataType* y6, DataType* x7,
                      DataType* y7, int* dim);

template <typename DataType>
void radix_prime_complexft(int n, int m, int p, int r, DataType* x, DataType* y,
                           int* dim);

/*     COMPLEX FINITE DISCRETE FOURIER TRANSFORM */
/*     TRANSFORMS ONE DIMENSION OF MULTI-DIMENSIONAL DATA */
/*     MODIFIED BY L. F. TEN EYCK FROM A ONE-DIMENSIONAL VERSION WRITTEN */
/*     BY G. T. SANDE, 1969. */
/*     THIS PROGRAM CALCULATES THE TRANSFORM */
/*               (X(T) + I*Y(T))*(COS(2*PI*T/N) - I*SIN(2*PI*T/N)) */

/*     INDEXING -- THE ARRANGEMENT OF THE MULTI-DIMENSIONAL DATA IS */
/*     SPECIFIED BY THE INTEGER ARRAY D, THE VALUES OF WHICH ARE USED AS */
/*     CONTROL PARAMETERS IN DO LOOPS.  WHEN IT IS DESIRED TO COVER ALL */
/*     ELEMENTS OF THE DATA FOR WHICH THE SUBSCRIPT BEING TRANSFORMED HAS */
/*     THE VALUE I0, THE FOLLOWING IS USED. */
/*               I1 = (I0 - 1)*D(2) + 1 */
/*               DO 100 I2 = I1, D(1), D(3) */
/*               I3 = I2 + D(4) - 1 */
/*               DO 100 I = I2, I3, D(5) */
/*                  . */
/*                  . */
/*           100 CONTINUE */
/*     WITH THIS INDEXING IT IS POSSIBLE TO USE A NUMBER OF ARRANGEMENTS */
/*     OF THE DATA, INCLUDING NORMAL FORTRAN COMPLEX NUMBERS (D(5) = 2) */
/*     OR SEPARATE STORAGE OF REAL AND IMAGINARY PARTS. */
/*     P MAX IS THE LARGEST PRIME FACTOR THAT WILL BE TOLERATED BY THIS */
/*     PROGRAM. */
/*     TWO GRP IS THE LARGEST POWER OF TWO THAT IS TREATED AS A SPECIAL CASE. */
template <typename DataType>
void complex_fourier_transform(DataType* x, DataType* y, int n, int* d) {

  constexpr int PrimeMax = 19;
  constexpr int TwoGrp = 8;

  int psym = 0;
  int unsym[15];
  int factors[15];
  int sym[15];

  if (n > 1) {
    symmetrized_reordering_factoring(n, PrimeMax, TwoGrp, factors, sym, &psym,
                                     unsym);
    complex_ft_kernel_driver(n, factors, d, x, y);
    double_inplace_reordering(n, sym, &psym, unsym, d, x, y);
  }
}

/*  SYMMETRIZED REORDERING FACTORING PROGRAM */
void symmetrized_reordering_factoring(int pts, int max_prime, int twogrp,
                                      int* factor, int* sym, int* psym,
                                      int* unsym) {

  int ptwo, j, r, jj, pp[14], qq[7];

  /* Parameter adjustments */
  --unsym;
  --sym;
  --factor;

  constexpr int nest = 14;

  int n = pts;
  *psym = 1;
  int f = 2;
  int p = 0;
  int q = 0;
L100:
  if (n <= 1) {
    goto L500;
  }
  for (j = f; j <= max_prime; ++j) {
    if (n == n / j * j) {
      goto L300;
    }
  }
  throw std::runtime_error("FFT: largest factor exceeds max_prime");
  return;
L300:
  if ((p << 1) + q >= nest) {
    throw std::runtime_error("FFT: factor count exceeds nest");
    return;
  }
  f = j;
  n /= f;
  if (n == n / f * f) {
    goto L400;
  }
  ++q;
  qq[q - 1] = f;
  goto L100;
L400:
  n /= f;
  ++p;
  pp[p - 1] = f;
  *psym *= f;
  goto L100;

L500:
  r = 1;
  if (q == 0) {
    r = 0;
  }
  if (p < 1) {
    goto L700;
  }
  max_prime = p;
#ifdef __INTEL_COMPILER
#pragma ivdep
#endif
  for (j = 1; j <= max_prime; ++j) {
    jj = p + 1 - j;
    sym[j] = pp[jj - 1];
    factor[j] = pp[jj - 1];
    jj = p + q + j;
    factor[jj] = pp[j - 1];
    jj = p + r + j;
    sym[jj] = pp[j - 1];
  }
L700:
  if (q < 1) {
    goto L900;
  }
  max_prime = q;
  for (j = 1; j <= max_prime; ++j) {
    jj = p + j;
    unsym[j] = qq[j - 1];
    factor[jj] = qq[j - 1];
  }
  max_prime = *psym;
  sym[p + 1] = pts / (max_prime * max_prime);
L900:
  jj = (p << 1) + q;
  factor[jj + 1] = 0;
  ptwo = 1;
  j = 0;
L1000:
  ++j;
  if (factor[j] == 0) {
    goto L1200;
  }
  if (factor[j] != 2) {
    goto L1000;
  }
  ptwo <<= 1;
  factor[j] = 1;
  if (ptwo >= twogrp) {
    goto L1100;
  }
  if (factor[j + 1] == 2) {
    goto L1000;
  }
L1100:
  factor[j] = ptwo;
  ptwo = 1;
  goto L1000;
L1200:
  if (p == 0) {
    r = 0;
  }
  jj = (p << 1) + r;
  sym[jj + 1] = 0;
  if (q <= 1) {
    q = 0;
  }
  unsym[q + 1] = 0;
}

/* DOUBLE IN PLACE REORDERING PROGRAM */
template <typename DataType>
void double_inplace_reordering(int pts, int* sym, int* psym, int* unsym,
                               int* dim, DataType* x, DataType* y) {
  /* System generated locals */
  int i__1, i__2, i__3, i__4, i__5, i__6, i__7, i__8, i__9, i__10, i__11, i__12,
      i__13, i__14, i__15, i__16, i__17, i__18, i__19, i__20, i__21, i__22,
      i__23, i__24, i__25, i__26, i__27, i__28, i__29, i__30, i__31;
  int equiv_25[14], equiv_26[14];

  /* Local variables */
  int mods = 0, size, test, mult, a, b, c, d, e, f, g, h, i, j, k, l, m, n, p;
#define s (equiv_25)

  DataType t;
#define u (equiv_26)

  int delta, p0, p1, p2, p3, p4, p5;
#define al (equiv_26)
#define bl (equiv_26 + 1)

  int dk;
#define cl (equiv_26 + 2)
#define dl (equiv_26 + 3)
#define el (equiv_26 + 4)
#define fl (equiv_26 + 5)

  int jj;
#define bs (equiv_25 + 1)

  int kk, lk;
#define cs (equiv_25 + 2)
#define ds (equiv_25 + 3)
#define es (equiv_25 + 4)
#define fs (equiv_25 + 5)
#define gs (equiv_25 + 6)
#define hs (equiv_25 + 7)
#define is (equiv_25 + 8)
#define js (equiv_25 + 9)
#define ks (equiv_25 + 10)
#define ls (equiv_25 + 11)

  int nt;
#define ms (equiv_25 + 12)
#define ns (equiv_25 + 13)
#define gl (equiv_26 + 6)
#define hl (equiv_26 + 7)
#define il (equiv_26 + 8)
#define jl (equiv_26 + 9)
#define kl (equiv_26 + 10)
#define ll (equiv_26 + 11)
#define ml (equiv_26 + 12)
#define nl (equiv_26 + 13)

  bool onemod;
  int modulo[14], punsym, sep;

  /* Parameter adjustments */
  --y;
  --x;
  --dim;
  --unsym;
  --sym;

  /* Function Body */
  const int nest = 14;

  nt = dim[1];
  sep = dim[2];
  p2 = dim[3];
  size = dim[4] - 1;
  p4 = dim[5];
  if (sym[1] == 0) {
    goto L500;
  }
  i__1 = nest;
  for (j = 1; j <= i__1; ++j) {
    u[j - 1] = 1;
    s[j - 1] = 1;
    /* L100: */
  }
  n = pts;
  i__1 = nest;
  for (j = 1; j <= i__1; ++j) {
    if (sym[j] == 0) {
      goto L300;
    }
    jj = nest + 1 - j;
    u[jj - 1] = n;
    s[jj - 1] = n / sym[j];
    n /= sym[j];
  }
L300:

  jj = 0;
  i__1 = *al;
  for (a = 1; a <= i__1; ++a) {
    i__2 = *bl;
    i__3 = *bs;
    for (b = a; b <= i__2; b += i__3) {
      i__4 = *cl;
      i__5 = *cs;
      for (c = b; c <= i__4; c += i__5) {
        i__6 = *dl;
        i__7 = *ds;
        for (d = c; d <= i__6; d += i__7) {
          i__8 = *el;
          i__9 = *es;
          for (e = d; e <= i__8; e += i__9) {

            i__10 = *fl;
            i__11 = *fs;
            for (f = e; f <= i__10; f += i__11) {
              i__12 = *gl;
              i__13 = *gs;
              for (g = f; g <= i__12; g += i__13) {
                i__14 = *hl;
                i__15 = *hs;
                for (h = g; h <= i__14; h += i__15) {
                  i__16 = *il;
                  i__17 = *is;
                  for (i = h; i <= i__16; i += i__17) {
                    i__18 = *jl;
                    i__19 = *js;
                    for (j = i; j <= i__18; j += i__19) {
                      i__20 = *kl;
                      i__21 = *ks;
                      for (k = j; k <= i__20; k += i__21) {
                        i__22 = *ll;
                        i__23 = *ls;
                        for (l = k; l <= i__22; l += i__23) {
                          i__24 = *ml;
                          i__25 = *ms;
                          for (m = l; m <= i__24; m += i__25) {
                            i__26 = *nl;
                            i__27 = *ns;
                            for (n = m; n <= i__26; n += i__27) {
                              ++jj;
                              if (jj >= n) {
                                goto L400;
                              }
                              delta = (n - jj) * sep;
                              p1 = (jj - 1) * sep + 1;
                              i__28 = nt;
                              i__29 = p2;
                              for (p0 = p1; p0 <= i__28; p0 += i__29) {
                                p3 = p0 + size;
                                i__30 = p3;
                                i__31 = p4;
                                for (p = p0; p <= i__30; p += i__31) {
                                  p5 = p + delta;
                                  t = x[p];
                                  x[p] = x[p5];
                                  x[p5] = t;
                                  t = y[p];
                                  y[p] = y[p5];
                                  y[p5] = t;
                                }
                              }
                            L400:
                              ;
                            }
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }

L500:
  if (unsym[1] == 0) {
    goto L1900;
  }
  /* Computing 2nd power */
  i__27 = *psym;
  punsym = pts / (i__27 * i__27);
  mult = punsym / unsym[1];
  test = (unsym[1] * unsym[2] - 1) * mult * *psym;
  lk = mult;
  dk = mult;
  i__27 = nest;
  for (k = 2; k <= i__27; ++k) {
    if (unsym[k] == 0) {
      goto L700;
    }
    lk *= unsym[k - 1];
    dk /= unsym[k];
    u[k - 1] = (lk - dk) * *psym;
    mods = k;
  }
L700:
  onemod = mods < 3;
  if (onemod) {
    goto L900;
  }
  i__27 = mods;
  for (j = 3; j <= i__27; ++j) {
    jj = mods + 3 - j;
    modulo[jj - 1] = u[j - 1];
  }
L900:
  modulo[1] = u[1];
  *jl = (punsym - 3) * *psym;
  *ms = punsym** psym;

  i__27 = *jl;
  i__26 = *psym;
  for (j = *psym; j <= i__27; j += i__26) {
    k = j;

  L1000:
    k *= mult;
    if (onemod) {
      goto L1200;
    }
    i__25 = mods;
    for (i = 3; i <= i__25; ++i) {
      k -= k / modulo[i - 1] * modulo[i - 1];
    }
  L1200:
    if (k >= test) {
      goto L1300;
    }
    k -= k / modulo[1] * modulo[1];
    goto L1400;
  L1300:
    k = k - k / modulo[1] * modulo[1] + modulo[1];
  L1400:
    if (k < j) {
      goto L1000;
    }

    if (k == j) {
      goto L1700;
    }
    delta = (k - j) * sep;
    i__25 = *psym;
    for (l = 1; l <= i__25; ++l) {
      i__24 = pts;
      i__23 = *ms;
      for (m = l; m <= i__24; m += i__23) {
        p1 = (m + j - 1) * sep + 1;
        i__22 = nt;
        i__21 = p2;
        for (p0 = p1; p0 <= i__22; p0 += i__21) {
          p3 = p0 + size;
          i__20 = p3;
          i__19 = p4;
          for (jj = p0; jj <= i__20; jj += i__19) {
            kk = jj + delta;
            t = x[jj];
            x[jj] = x[kk];
            x[kk] = t;
            t = y[jj];
            y[jj] = y[kk];
            y[kk] = t;
          }
        }
      }
    }
  L1700:
    ;
  }

L1900:
  return;
}

#undef nl
#undef ml
#undef ll
#undef kl
#undef jl
#undef il
#undef hl
#undef gl
#undef ns
#undef ms
#undef ls
#undef ks
#undef js
#undef is
#undef hs
#undef gs
#undef fs
#undef es
#undef ds
#undef cs
#undef bs
#undef fl
#undef el
#undef dl
#undef cl
#undef bl
#undef al
#undef u
#undef s

/* MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL DRIVER */
template <typename DataType>
void complex_ft_kernel_driver(int n, int* factor, int* dim, DataType* x,
                              DataType* y) {

  int f = -1;
  int m = n;
  int p = 0;
  while ((p = factor[++f]) > 0) {
    m /= p;
    int r = m * dim[1];
    switch (p) {
      case 1:
        break;
      case 2:
        radix2_complexft(n, m, x, y, &x[r], &y[r], dim);
        break;
      case 3:
        radix3_complexft(n, m, x, y, &x[r], &y[r], &x[r * 2], &y[r * 2], dim);
        break;
      case 4:
        radix4_complexft(n, m, x, y, &x[r], &y[r], &x[r * 2], &y[r * 2],
                         &x[r * 3], &y[r * 3], dim);
        break;
      case 5:
        radix5_complexft(n, m, x, y, &x[r], &y[r], &x[r * 2], &y[r * 2],
                         &x[r * 3], &y[r * 3], &x[r * 4], &y[r * 4], dim);
        break;
      case 6:
        throw std::runtime_error(
            "FFT: Bad factor (6) in ComplexFTKernalDriver");
      case 8:
        radix8_complexft(n, m, x, y, &x[r], &y[r], &x[r * 2], &y[r * 2],
                         &x[r * 3], &y[r * 3], &x[r * 4], &y[r * 4], &x[r * 5],
                         &y[r * 5], &x[r * 6], &y[r * 6], &x[r * 7], &y[r * 7],
                         dim);
        break;
      default:
        radix_prime_complexft(n, m, p, r, x, y, dim);
        break;
    }
  }
}

/* RADIX 2 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL */
template <typename DataType>
void radix2_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, int* dim) {
  /* System generated locals */
  int i__1, i__2, i__3, i__4, i__5, i__6, i__7;

  /* Local variables */
  bool fold = false;
  int size = 0;
  bool zero = false;
  DataType c = 0;
  int j = 0, k = 0, l = 0;
  DataType s = 0, angle = 0;
  int k0 = 0, k1 = 0, k2 = 0, l1 = 0, m2 = 0, mover2 = 0, kk = 0;
  DataType is = 0, iu = 0;
  int ns = 0, nt = 0;
  DataType rs = 0, ru = 0, fm2 = 0;
  int mm2, sep;
  DataType fjm1;

  /* Parameter adjustments */
  --dim;
  --y1;
  --x1;
  --y0;
  --x0;

  /* Function Body */

  nt = dim[1];
  sep = dim[2];
  l1 = dim[3];
  size = dim[4] - 1;
  k2 = dim[5];
  ns = n * sep;
  m2 = m << 1;
  fm2 = (DataType)m2;
  mover2 = m / 2 + 1;
  mm2 = sep * m2;

  fjm1 = (DataType)-1.0;
  i__1 = mover2;
  for (j = 1; j <= i__1; ++j) {
    fold = j > 1 && j << 1 < m + 2;
    k0 = (j - 1) * sep + 1;
    fjm1 += (DataType)1.0;
    angle = (DataType)(2.0 * geom::PI * fjm1 / fm2);
    zero = (angle == 0.0);
    if (!zero) {
      c = std::cos(angle);
      s = std::sin(angle);
    }
    goto L200;
  L100:
    fold = false;
    k0 = (m + 1 - j) * sep + 1;
    c = -c;
  L200:
    i__2 = ns;
    i__3 = mm2;
    for (kk = k0; kk <= i__2; kk += i__3) {
      i__4 = nt;
      i__5 = l1;
      for (l = kk; l <= i__4; l += i__5) {
        k1 = l + size;
        i__6 = k1;
        i__7 = k2;
#ifdef __INTEL_COMPILER
#pragma ivdep
#endif
        for (k = l; k <= i__6; k += i__7) {
          rs = x0[k] + x1[k];
          is = y0[k] + y1[k];
          ru = x0[k] - x1[k];
          iu = y0[k] - y1[k];
          x0[k] = rs;
          y0[k] = is;
          if (zero) {
            x1[k] = ru;
            y1[k] = iu;
          } else {
            x1[k] = ru * c + iu * s;
            y1[k] = iu * c - ru * s;
          }
        }
      }
    }
    if (fold) {
      goto L100;
    }
  }
  return;
}

template <typename DataType>
void radix3_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, int* dim) {

  const DataType a = (DataType)-0.5;
  const DataType b = (DataType)0.866025403784438647; // sqrt(3)/2

  /* System generated locals */
  int i__1, i__2, i__3, i__4, i__5, i__6, i__7;

  /* Local variables */
  bool fold = false;
  int size = 0;
  bool zero = false;
  int j = 0, k = 0, l = 0;
  DataType t = 0, angle = 0, c1 = 0, c2 = 0, i0 = 0, i1 = 0;
  int k0 = 0, k1 = 0, k2 = 0, l1 = 0;
  DataType i2 = 0;
  int m3 = 0;
  DataType r0 = 0, s1 = 0, s2 = 0, r1 = 0, r2 = 0;
  int mover2 = 0;
  DataType ia = 0, ib = 0, ra = 0, rb = 0;
  int kk = 0;
  DataType is = 0;
  int ns = 0, nt = 0;
  DataType rs = 0, fm3 = 0;
  int mm3 = 0, sep = 0;
  DataType fjm1 = 0;

  /* RADIX 3 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL */

  /* Parameter adjustments */
  --dim;
  --y2;
  --x2;
  --y1;
  --x1;
  --y0;
  --x0;

  /* Function Body */

  nt = dim[1];
  sep = dim[2];
  l1 = dim[3];
  size = dim[4] - 1;
  k2 = dim[5];
  ns = n * sep;
  m3 = m * 3;
  fm3 = (DataType)m3;
  mm3 = sep * m3;
  mover2 = m / 2 + 1;

  fjm1 = (DataType)-1.0;
  i__1 = mover2;
  for (j = 1; j <= i__1; ++j) {
    fold = j > 1 && j << 1 < m + 2;
    k0 = (j - 1) * sep + 1;
    fjm1 += (DataType)1.0;
    angle = (DataType)(2.0 * geom::PI * fjm1 / fm3);
    zero = angle == 0.0;
    if (zero) {
      goto L200;
    }
    c1 = std::cos(angle);
    s1 = std::sin(angle);
    c2 = c1 * c1 - s1 * s1;
    s2 = s1 * c1 + c1 * s1;
    goto L200;
  L100:
    fold = false;
    k0 = (m + 1 - j) * sep + 1;
    t = c1 * a + s1 * b;
    s1 = c1 * b - s1 * a;
    c1 = t;
    t = c2 * a - s2 * b;
    s2 = -c2 * b - s2 * a;
    c2 = t;
  L200:

    i__2 = ns;
    i__3 = mm3;
    for (kk = k0; kk <= i__2; kk += i__3) {
      i__4 = nt;
      i__5 = l1;
      for (l = kk; l <= i__4; l += i__5) {
        k1 = l + size;
        i__6 = k1;
        i__7 = k2;
#ifdef __INTEL_COMPILER
#pragma ivdep
#endif
        for (k = l; k <= i__6; k += i__7) {
          r0 = x0[k];
          i0 = y0[k];
          rs = x1[k] + x2[k];
          is = y1[k] + y2[k];
          x0[k] = r0 + rs;
          y0[k] = i0 + is;
          ra = r0 + rs * a;
          ia = i0 + is * a;
          rb = (x1[k] - x2[k]) * b;
          ib = (y1[k] - y2[k]) * b;
          if (zero) {
            x1[k] = ra + ib;
            y1[k] = ia - rb;
            x2[k] = ra - ib;
            y2[k] = ia + rb;
          } else {
            r1 = ra + ib;
            i1 = ia - rb;
            r2 = ra - ib;
            i2 = ia + rb;
            x1[k] = r1 * c1 + i1 * s1;
            y1[k] = i1 * c1 - r1 * s1;
            x2[k] = r2 * c2 + i2 * s2;
            y2[k] = i2 * c2 - r2 * s2;
          }
        }
      }
    }
    if (fold) {
      goto L100;
    }
  }

  return;
}

/* RADIX 4 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL */
template <typename DataType>
void radix4_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, DataType* x3,
                      DataType* y3, int* dim) {
  /* System generated locals */
  int i__1 = 0, i__2 = 0, i__3 = 0, i__4 = 0, i__5 = 0, i__6 = 0, i__7 = 0;

  /* Local variables */
  bool fold = false;
  int size = 0;
  bool zero = false;
  int j = 0, k = 0, l = 0;
  DataType t = 0, angle = 0, c1 = 0, c2 = 0, c3 = 0, i1 = 0;
  int k0 = 0, k1 = 0, k2 = 0, l1 = 0;
  DataType i2 = 0, i3 = 0;
  int m4 = 0;
  DataType r1 = 0, s1 = 0, s2 = 0, s3 = 0, r2 = 0, r3 = 0;
  int mover2 = 0, kk = 0, ns = 0, nt = 0;
  DataType fm4 = 0, is0 = 0, is1 = 0;
  int mm4 = 0;
  DataType iu0 = 0, iu1 = 0, rs0 = 0, rs1 = 0, ru0 = 0, ru1 = 0;
  int sep = 0;
  DataType fjm1 = 0;

  /* Parameter adjustments */
  --dim;
  --y3;
  --x3;
  --y2;
  --x2;
  --y1;
  --x1;
  --y0;
  --x0;

  /* Function Body */

  nt = dim[1];
  sep = dim[2];
  l1 = dim[3];
  size = dim[4] - 1;
  k2 = dim[5];
  ns = n * sep;
  m4 = m << 2;
  fm4 = (DataType)m4;
  mm4 = sep * m4;
  mover2 = m / 2 + 1;

  fjm1 = (DataType)-1.0;
  i__1 = mover2;
  for (j = 1; j <= i__1; ++j) {
    fold = j > 1 && j << 1 < m + 2;
    k0 = (j - 1) * sep + 1;
    fjm1 += (DataType)1.0;
    angle = (DataType)(2.0 * geom::PI * fjm1 / fm4);
    zero = angle == 0.0;
    if (zero) {
      goto L200;
    }
    c1 = std::cos(angle);
    s1 = std::sin(angle);
    c2 = c1 * c1 - s1 * s1;
    s2 = s1 * c1 + c1 * s1;
    c3 = c2 * c1 - s2 * s1;
    s3 = s2 * c1 + c2 * s1;
    goto L200;
  L100:
    fold = false;
    k0 = (m + 1 - j) * sep + 1;
    t = c1;
    c1 = s1;
    s1 = t;
    c2 = -c2;
    t = c3;
    c3 = -s3;
    s3 = -t;
  L200:

    i__2 = ns;
    i__3 = mm4;
#ifdef __INTEL_COMPILER
#pragma ivdep
#endif
    for (kk = k0; kk <= i__2; kk += i__3) {
      i__4 = nt;
      i__5 = l1;
      for (l = kk; l <= i__4; l += i__5) {
        k1 = l + size;
        i__6 = k1;
        i__7 = k2;
#ifdef __INTEL_COMPILER
#pragma ivdep
#endif
        for (k = l; k <= i__6; k += i__7) {
          rs0 = x0[k] + x2[k];
          is0 = y0[k] + y2[k];
          ru0 = x0[k] - x2[k];
          iu0 = y0[k] - y2[k];
          rs1 = x1[k] + x3[k];
          is1 = y1[k] + y3[k];
          ru1 = x1[k] - x3[k];
          iu1 = y1[k] - y3[k];
          x0[k] = rs0 + rs1;
          y0[k] = is0 + is1;
          if (zero) {
            goto L300;
          }
          r1 = ru0 + iu1;
          i1 = iu0 - ru1;
          r2 = rs0 - rs1;
          i2 = is0 - is1;
          r3 = ru0 - iu1;
          i3 = iu0 + ru1;
          x2[k] = r1 * c1 + i1 * s1;
          y2[k] = i1 * c1 - r1 * s1;
          x1[k] = r2 * c2 + i2 * s2;
          y1[k] = i2 * c2 - r2 * s2;
          x3[k] = r3 * c3 + i3 * s3;
          y3[k] = i3 * c3 - r3 * s3;
          goto L400;
        L300:
          x2[k] = ru0 + iu1;
          y2[k] = iu0 - ru1;
          x1[k] = rs0 - rs1;
          y1[k] = is0 - is1;
          x3[k] = ru0 - iu1;
          y3[k] = iu0 + ru1;
        L400:
          ;
        }
      }
    }
    if (fold) {
      goto L100;
    }
  }

  return;
}

/* RADIX 5 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL */
template <typename DataType>
void radix5_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, DataType* x3,
                      DataType* y3, DataType* x4, DataType* y4, int* dim) {
  /* Initialized data */

  const DataType a1 = (DataType)0.30901699;
  const DataType b1 = (DataType)0.95105652;
  const DataType a2 = (DataType)-0.80901699;
  const DataType b2 = (DataType)0.58778525;

  /* System generated locals */
  int i__1, i__2, i__3, i__4, i__5, i__6, i__7;

  /* Local variables */
  bool fold = false;
  int size = 0;
  bool zero = false;
  int j = 0, k = 0, l = 0;
  DataType t = 0, angle = 0, c1 = 0, c2 = 0, c3 = 0, c4 = 0, i0 = 0;
  int k0 = 0, k1 = 0, k2 = 0, l1 = 0;
  DataType i1 = 0, i2 = 0, i3 = 0;
  int m5 = 0;
  DataType s1 = 0, s2 = 0, s3 = 0, s4 = 0, r0 = 0, r1 = 0, r2 = 0, r3 = 0,
           r4 = 0, i4 = 0;
  int mover2 = 0, kk = 0, ns = 0, nt = 0;
  DataType ia1 = 0, ia2 = 0, ib1 = 0, ib2 = 0, ra1 = 0, ra2 = 0, rb1 = 0,
           rb2 = 0, fm5 = 0, is1 = 0, is2 = 0;
  int mm5 = 0;
  DataType iu1 = 0, iu2 = 0, rs1 = 0, rs2 = 0, ru1 = 0, ru2 = 0;
  int sep = 0;
  DataType fjm1 = 0;

  /* Parameter adjustments */
  --dim;
  --y4;
  --x4;
  --y3;
  --x3;
  --y2;
  --x2;
  --y1;
  --x1;
  --y0;
  --x0;

  /* Function Body */

  nt = dim[1];
  sep = dim[2];
  l1 = dim[3];
  size = dim[4] - 1;
  k2 = dim[5];
  ns = n * sep;
  m5 = m * 5;
  fm5 = (DataType)m5;
  mm5 = sep * m5;
  mover2 = m / 2 + 1;

  fjm1 = (DataType)-1.0;
  i__1 = mover2;
  for (j = 1; j <= i__1; ++j) {
    fold = j > 1 && j << 1 < m + 2;
    k0 = (j - 1) * sep + 1;
    fjm1 += (DataType)1.0;
    angle = (DataType)(2.0 * geom::PI * fjm1 / fm5);
    zero = angle == 0.0;
    if (zero) {
      goto L200;
    }
    c1 = std::cos(angle);
    s1 = std::sin(angle);
    c2 = c1 * c1 - s1 * s1;
    s2 = s1 * c1 + c1 * s1;
    c3 = c2 * c1 - s2 * s1;
    s3 = s2 * c1 + c2 * s1;
    c4 = c2 * c2 - s2 * s2;
    s4 = s2 * c2 + c2 * s2;
    goto L200;
  L100:
    fold = false;
    k0 = (m + 1 - j) * sep + 1;
    t = c1 * a1 + s1 * b1;
    s1 = c1 * b1 - s1 * a1;
    c1 = t;
    t = c2 * a2 + s2 * b2;
    s2 = c2 * b2 - s2 * a2;
    c2 = t;
    t = c3 * a2 - s3 * b2;
    s3 = -c3 * b2 - s3 * a2;
    c3 = t;
    t = c4 * a1 - s4 * b1;
    s4 = -c4 * b1 - s4 * a1;
    c4 = t;
  L200:

    i__2 = ns;
    i__3 = mm5;
    for (kk = k0; kk <= i__2; kk += i__3) {
      i__4 = nt;
      i__5 = l1;
      for (l = kk; l <= i__4; l += i__5) {
        k1 = l + size;
        i__6 = k1;
        i__7 = k2;
        for (k = l; k <= i__6; k += i__7) {
          r0 = x0[k];
          i0 = y0[k];
          rs1 = x1[k] + x4[k];
          is1 = y1[k] + y4[k];
          ru1 = x1[k] - x4[k];
          iu1 = y1[k] - y4[k];
          rs2 = x2[k] + x3[k];
          is2 = y2[k] + y3[k];
          ru2 = x2[k] - x3[k];
          iu2 = y2[k] - y3[k];
          x0[k] = r0 + rs1 + rs2;
          y0[k] = i0 + is1 + is2;
          ra1 = r0 + rs1 * a1 + rs2 * a2;
          ia1 = i0 + is1 * a1 + is2 * a2;
          ra2 = r0 + rs1 * a2 + rs2 * a1;
          ia2 = i0 + is1 * a2 + is2 * a1;
          rb1 = ru1 * b1 + ru2 * b2;
          ib1 = iu1 * b1 + iu2 * b2;
          rb2 = ru1 * b2 - ru2 * b1;
          ib2 = iu1 * b2 - iu2 * b1;
          if (zero) {
            goto L300;
          }
          r1 = ra1 + ib1;
          i1 = ia1 - rb1;
          r2 = ra2 + ib2;
          i2 = ia2 - rb2;
          r3 = ra2 - ib2;
          i3 = ia2 + rb2;
          r4 = ra1 - ib1;
          i4 = ia1 + rb1;
          x1[k] = r1 * c1 + i1 * s1;
          y1[k] = i1 * c1 - r1 * s1;
          x2[k] = r2 * c2 + i2 * s2;
          y2[k] = i2 * c2 - r2 * s2;
          x3[k] = r3 * c3 + i3 * s3;
          y3[k] = i3 * c3 - r3 * s3;
          x4[k] = r4 * c4 + i4 * s4;
          y4[k] = i4 * c4 - r4 * s4;
          goto L400;
        L300:
          x1[k] = ra1 + ib1;
          y1[k] = ia1 - rb1;
          x2[k] = ra2 + ib2;
          y2[k] = ia2 - rb2;
          x3[k] = ra2 - ib2;
          y3[k] = ia2 + rb2;
          x4[k] = ra1 - ib1;
          y4[k] = ia1 + rb1;
        L400:
          ;
        }
      }
    }
    if (fold) {
      goto L100;
    }
  }
  return;
}

/* RADIX 8 MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL */
template <typename DataType>
void radix8_complexft(int n, int m, DataType* x0, DataType* y0, DataType* x1,
                      DataType* y1, DataType* x2, DataType* y2, DataType* x3,
                      DataType* y3, DataType* x4, DataType* y4, DataType* x5,
                      DataType* y5, DataType* x6, DataType* y6, DataType* x7,
                      DataType* y7, int* dim) {

  const DataType e = (DataType)0.707106781186547524; /* sqrt(2)*0.5; */

  /* System generated locals */
  int i__1, i__2, i__3, i__4, i__5, i__6, i__7;

  /* Local variables */
  bool fold = false;
  int size = 0;
  bool zero = false;
  int j = 0, k = 0, l = 0;
  DataType t = 0, angle = 0, c1 = 0, c2 = 0, c3 = 0, c4 = 0, c5 = 0, c6 = 0,
           c7 = 0;
  int k0 = 0, k1 = 0, k2 = 0, l1 = 0;
  DataType i1 = 0, i2 = 0, i3 = 0, r1 = 0, s1 = 0;
  int m8 = 0;
  DataType s2 = 0, s3 = 0, s4 = 0, s5 = 0, s6 = 0, s7 = 0, r2 = 0, r3 = 0,
           r4 = 0, r5 = 0, r6 = 0, r7 = 0, i4 = 0, i5 = 0, i6 = 0, i7 = 0;
  int mover2 = 0, kk = 0, ns = 0, nt = 0;
  DataType fm8 = 0, is0 = 0, is1 = 0, is2 = 0, is3 = 0, iu0 = 0, iu1 = 0;
  int mm8 = 0;
  DataType iu2 = 0, iu3 = 0, rs0 = 0, rs1 = 0, rs2 = 0, rs3 = 0, ru0 = 0,
           ru1 = 0, ru2 = 0, ru3 = 0;
  int sep = 0;
  DataType fjm1 = 0, iss0 = 0, iss1 = 0, isu0 = 0, isu1 = 0, ius0 = 0, ius1 = 0,
           iuu0 = 0, iuu1 = 0, rss0 = 0, rss1 = 0, rsu0 = 0, rsu1 = 0, rus0 = 0,
           rus1 = 0, ruu0 = 0, ruu1 = 0;

  /* Parameter adjustments */
  --dim;
  --y7;
  --x7;
  --y6;
  --x6;
  --y5;
  --x5;
  --y4;
  --x4;
  --y3;
  --x3;
  --y2;
  --x2;
  --y1;
  --x1;
  --y0;
  --x0;

  /* Function Body */

  nt = dim[1];
  sep = dim[2];
  l1 = dim[3];
  size = dim[4] - 1;
  k2 = dim[5];
  ns = n * sep;
  m8 = m << 3;
  fm8 = (DataType)m8;
  mm8 = sep * m8;
  mover2 = m / 2 + 1;

  fjm1 = (DataType)-1.0;
  i__1 = mover2;
  for (j = 1; j <= i__1; ++j) {
    fold = j > 1 && j << 1 < m + 2;
    k0 = (j - 1) * sep + 1;
    fjm1 += (DataType)1.0;
    angle = (DataType)(2.0 * geom::PI * fjm1 / fm8);
    zero = angle == 0.0;
    if (zero) {
      goto L200;
    }
    c1 = std::cos(angle);
    s1 = std::sin(angle);
    c2 = c1 * c1 - s1 * s1;
    s2 = s1 * c1 + c1 * s1;
    c3 = c2 * c1 - s2 * s1;
    s3 = s2 * c1 + c2 * s1;
    c4 = c2 * c2 - s2 * s2;
    s4 = s2 * c2 + c2 * s2;
    c5 = c4 * c1 - s4 * s1;
    s5 = s4 * c1 + c4 * s1;
    c6 = c4 * c2 - s4 * s2;
    s6 = s4 * c2 + c4 * s2;
    c7 = c4 * c3 - s4 * s3;
    s7 = s4 * c3 + c4 * s3;
    goto L200;
  L100:
    fold = false;
    k0 = (m + 1 - j) * sep + 1;
    t = (c1 + s1) * e;
    s1 = (c1 - s1) * e;
    c1 = t;
    t = s2;
    s2 = c2;
    c2 = t;
    t = (-c3 + s3) * e;
    s3 = (c3 + s3) * e;
    c3 = t;
    c4 = -c4;
    t = -(c5 + s5) * e;
    s5 = (-c5 + s5) * e;
    c5 = t;
    t = -s6;
    s6 = -c6;
    c6 = t;
    t = (c7 - s7) * e;
    s7 = -(c7 + s7) * e;
    c7 = t;
  L200:

    i__2 = ns;
    i__3 = mm8;
    for (kk = k0; kk <= i__2; kk += i__3) {
      i__4 = nt;
      i__5 = l1;
      for (l = kk; l <= i__4; l += i__5) {
        k1 = l + size;
        i__6 = k1;
        i__7 = k2;
        for (k = l; k <= i__6; k += i__7) {
          rs0 = x0[k] + x4[k];
          is0 = y0[k] + y4[k];
          ru0 = x0[k] - x4[k];
          iu0 = y0[k] - y4[k];
          rs1 = x1[k] + x5[k];
          is1 = y1[k] + y5[k];
          ru1 = x1[k] - x5[k];
          iu1 = y1[k] - y5[k];
          rs2 = x2[k] + x6[k];
          is2 = y2[k] + y6[k];
          ru2 = x2[k] - x6[k];
          iu2 = y2[k] - y6[k];
          rs3 = x3[k] + x7[k];
          is3 = y3[k] + y7[k];
          ru3 = x3[k] - x7[k];
          iu3 = y3[k] - y7[k];
          rss0 = rs0 + rs2;
          iss0 = is0 + is2;
          rsu0 = rs0 - rs2;
          isu0 = is0 - is2;
          rss1 = rs1 + rs3;
          iss1 = is1 + is3;
          rsu1 = rs1 - rs3;
          isu1 = is1 - is3;
          rus0 = ru0 - iu2;
          ius0 = iu0 + ru2;
          ruu0 = ru0 + iu2;
          iuu0 = iu0 - ru2;
          rus1 = ru1 - iu3;
          ius1 = iu1 + ru3;
          ruu1 = ru1 + iu3;
          iuu1 = iu1 - ru3;
          t = (rus1 + ius1) * e;
          ius1 = (ius1 - rus1) * e;
          rus1 = t;
          t = (ruu1 + iuu1) * e;
          iuu1 = (iuu1 - ruu1) * e;
          ruu1 = t;
          x0[k] = rss0 + rss1;
          y0[k] = iss0 + iss1;
          if (zero) {
            x4[k] = ruu0 + ruu1;
            y4[k] = iuu0 + iuu1;
            x2[k] = rsu0 + isu1;
            y2[k] = isu0 - rsu1;
            x6[k] = rus0 + ius1;
            y6[k] = ius0 - rus1;
            x1[k] = rss0 - rss1;
            y1[k] = iss0 - iss1;
            x5[k] = ruu0 - ruu1;
            y5[k] = iuu0 - iuu1;
            x3[k] = rsu0 - isu1;
            y3[k] = isu0 + rsu1;
            x7[k] = rus0 - ius1;
            y7[k] = ius0 + rus1;
          } else {
            r1 = ruu0 + ruu1;
            i1 = iuu0 + iuu1;
            r2 = rsu0 + isu1;
            i2 = isu0 - rsu1;
            r3 = rus0 + ius1;
            i3 = ius0 - rus1;
            r4 = rss0 - rss1;
            i4 = iss0 - iss1;
            r5 = ruu0 - ruu1;
            i5 = iuu0 - iuu1;
            r6 = rsu0 - isu1;
            i6 = isu0 + rsu1;
            r7 = rus0 - ius1;
            i7 = ius0 + rus1;
            x4[k] = r1 * c1 + i1 * s1;
            y4[k] = i1 * c1 - r1 * s1;
            x2[k] = r2 * c2 + i2 * s2;
            y2[k] = i2 * c2 - r2 * s2;
            x6[k] = r3 * c3 + i3 * s3;
            y6[k] = i3 * c3 - r3 * s3;
            x1[k] = r4 * c4 + i4 * s4;
            y1[k] = i4 * c4 - r4 * s4;
            x5[k] = r5 * c5 + i5 * s5;
            y5[k] = i5 * c5 - r5 * s5;
            x3[k] = r6 * c6 + i6 * s6;
            y3[k] = i6 * c6 - r6 * s6;
            x7[k] = r7 * c7 + i7 * s7;
            y7[k] = i7 * c7 - r7 * s7;
          }
        }
      }
    }
    if (fold) {
      goto L100;
    }
  }
}

/* RADIX PRIME MULTI-DIMENSIONAL COMPLEX FOURIER TRANSFORM KERNEL */
template <typename DataType>
void radix_prime_complexft(int n, int m, int p, int r, DataType* x, DataType* y,
                           int* dim) {
  /* System generated locals */
  int x_dim1 = 0, x_offset = 0, y_dim1 = 0, y_offset = 0, i__1 = 0, i__2 = 0,
      i__3 = 0, i__4 = 0, i__5 = 0, i__6 = 0, i__7 = 0, i__8 = 0, i__9 = 0;

  /* Local variables */
  bool fold = false;
  int size = 0;
  bool zero = false;
  DataType a[18], b[18], c[18];
  int j = 0, k = 0, l = 0;
  DataType s[18], t = 0;
  int u = 0;
  DataType angle = 0;
  int v = 0, k0 = 0, k1 = 0, k2 = 0, l1 = 0;
  DataType aa[81], bb[81];
  int mover2 = 0;
  DataType ia[9], ib[9], ra[9];
  int jj = 0;
  DataType rb[9], fp;
  int kk = 0;
  DataType fu, is;
  int mp = 0;
  DataType iu = 0;
  int pm = 0, pp = 0, ns = 0, nt = 0;
  DataType rs = 0, ru = 0, xt = 0, yt = 0, fmp = 0;
  int sep = 0, mmp = 0;
  DataType fjm1 = 0;

  /* Parameter adjustments */
  --dim;
  y_dim1 = r;
  y_offset = y_dim1 + 1;
  y -= y_offset;
  x_dim1 = r;
  x_offset = x_dim1 + 1;
  x -= x_offset;

  /* Function Body */

  nt = dim[1];
  sep = dim[2];
  l1 = dim[3];
  size = dim[4] - 1;
  k2 = dim[5];
  ns = n * sep;
  mover2 = m / 2 + 1;
  mp = m * p;
  fmp = (DataType)mp;
  mmp = sep * mp;
  pp = p / 2;
  pm = p - 1;
  fp = (DataType)(p);
  fu = (DataType)0.0;
  i__1 = pp;
#ifdef __INTEL_COMPILER
#pragma ivdep
#endif
  for (u = 1; u <= i__1; ++u) {
    fu += (DataType)1.0;
    angle = (DataType)(2.0 * geom::PI * fu / fp);
    jj = p - u;
    a[u - 1] = std::cos(angle);
    b[u - 1] = std::sin(angle);
    a[jj - 1] = a[u - 1];
    b[jj - 1] = -b[u - 1];
  }
  i__1 = pp;
  for (u = 1; u <= i__1; ++u) {
    i__2 = pp;
    for (v = 1; v <= i__2; ++v) {
      jj = u * v - u * v / p * p;
      aa[v + u * 9 - 10] = a[jj - 1];
      bb[v + u * 9 - 10] = b[jj - 1];
    }
  }

  fjm1 = (DataType)-1.0;
  i__1 = mover2;
  for (j = 1; j <= i__1; ++j) {
    fold = j > 1 && j << 1 < m + 2;
    k0 = (j - 1) * sep + 1;
    fjm1 += (DataType)1.0;
    angle = (DataType)(2.0 * geom::PI * fjm1 / fmp);
    zero = (angle == 0.0);
    if (zero) {
      goto L700;
    }
    c[0] = std::cos(angle);
    s[0] = std::sin(angle);
    i__2 = pm;
    for (u = 2; u <= i__2; ++u) {
      c[u - 1] = c[u - 2] * c[0] - s[u - 2] * s[0];
      s[u - 1] = s[u - 2] * c[0] + c[u - 2] * s[0];
    }
    goto L700;
  L500:
    fold = false;
    k0 = (m + 1 - j) * sep + 1;
    i__2 = pm;
    for (u = 1; u <= i__2; ++u) {
      t = c[u - 1] * a[u - 1] + s[u - 1] * b[u - 1];
      s[u - 1] = -s[u - 1] * a[u - 1] + c[u - 1] * b[u - 1];
      c[u - 1] = t;
    }
  L700:

    i__2 = ns;
    i__3 = mmp;
    for (kk = k0; kk <= i__2; kk += i__3) {
      i__4 = nt;
      i__5 = l1;
      for (l = kk; l <= i__4; l += i__5) {
        k1 = l + size;
        i__6 = k1;
        i__7 = k2;
        for (k = l; k <= i__6; k += i__7) {
          xt = x[k + x_dim1];
          yt = y[k + y_dim1];
          rs = x[k + (x_dim1 << 1)] + x[k + p * x_dim1];
          is = y[k + (y_dim1 << 1)] + y[k + p * y_dim1];
          ru = x[k + (x_dim1 << 1)] - x[k + p * x_dim1];
          iu = y[k + (y_dim1 << 1)] - y[k + p * y_dim1];
          i__8 = pp;
          for (u = 1; u <= i__8; ++u) {
            ra[u - 1] = xt + rs * aa[u - 1];
            ia[u - 1] = yt + is * aa[u - 1];
            rb[u - 1] = ru * bb[u - 1];
            ib[u - 1] = iu * bb[u - 1];
          }
          xt += rs;
          yt += is;
          i__8 = pp;
          for (u = 2; u <= i__8; ++u) {
            jj = p - u;
            rs = x[k + (u + 1) * x_dim1] + x[k + (jj + 1) * x_dim1];
            is = y[k + (u + 1) * y_dim1] + y[k + (jj + 1) * y_dim1];
            ru = x[k + (u + 1) * x_dim1] - x[k + (jj + 1) * x_dim1];
            iu = y[k + (u + 1) * y_dim1] - y[k + (jj + 1) * y_dim1];
            xt += rs;
            yt += is;
            i__9 = pp;
            for (v = 1; v <= i__9; ++v) {
              ra[v - 1] += rs * aa[v + u * 9 - 10];
              ia[v - 1] += is * aa[v + u * 9 - 10];
              rb[v - 1] += ru * bb[v + u * 9 - 10];
              ib[v - 1] += iu * bb[v + u * 9 - 10];
            }
          }
          x[k + x_dim1] = xt;
          y[k + y_dim1] = yt;
          i__8 = pp;
#ifdef __INTEL_COMPILER
#pragma ivdep
#endif
          for (u = 1; u <= i__8; ++u) {
            jj = p - u;
            if (zero) {
              x[k + (u + 1) * x_dim1] = ra[u - 1] + ib[u - 1];
              y[k + (u + 1) * y_dim1] = ia[u - 1] - rb[u - 1];
              x[k + (jj + 1) * x_dim1] = ra[u - 1] - ib[u - 1];
              y[k + (jj + 1) * y_dim1] = ia[u - 1] + rb[u - 1];
            } else {
              xt = ra[u - 1] + ib[u - 1];
              yt = ia[u - 1] - rb[u - 1];
              x[k + (u + 1) * x_dim1] = xt * c[u - 1] + yt * s[u - 1];
              y[k + (u + 1) * y_dim1] = yt * c[u - 1] - xt * s[u - 1];
              xt = ra[u - 1] - ib[u - 1];
              yt = ia[u - 1] + rb[u - 1];
              x[k + (jj + 1) * x_dim1] = xt * c[jj - 1] + yt * s[jj - 1];
              y[k + (jj + 1) * y_dim1] = yt * c[jj - 1] - xt * s[jj - 1];
            }
          }
        }
      }
    }
    if (fold) {
      goto L500;
    }
  }
}

#endif
