/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef MOLECULARTRANSFORM_HPP
#define MOLECULARTRANSFORM_HPP

#include "grid3.hpp"
#include "atom.hpp"
#include "fouriermap.hpp"
#include "unitcell.hpp"
#include "geometry.hpp"
#include <complex>
#include <vector>
#include <array>
#include <limits>

namespace crystallography {

template <class T>
class molecular_transform {
public:
  molecular_transform(const std::vector<atom>& atoms, double resolution_limit,
                      double box_scale, double fft_sampling);

  ~molecular_transform();

  std::complex<T> structure_factor(double h, double k, double l) const {
    return sfgrid->linear_interpolated_value(h, k, l);
  }

  template <class RealT>
  std::complex<T> structure_factor(const std::array<RealT, 3>& hkl) const {
    return sfgrid->linear_interpolated_value(hkl);
  }

  std::complex<T> nearest_integral_structure_factor(double h, double k,
                                                    double l) const {
    return sfgrid->nearest_value(h, k, l);
  }

  template <class RealT>
  std::complex<T>
  nearest_integral_structure_factor(const std::array<RealT, 3>& hkl) const {
    return sfgrid->nearest_value(hkl);
  }

  unit_cell& unitcell() const { return *cell; }

  const grid3<std::complex<T>>& reflection_grid() const { return *sfgrid; }

protected:
  unit_cell* cell;
  grid3<std::complex<T>>* sfgrid;
};

template <class T>
molecular_transform<T>::molecular_transform(const std::vector<atom>& atoms,
                                            double resolution_limit,
                                            double box_scale,
                                            double fft_sampling) {
  auto low = std::array<double, 3>{{std::numeric_limits<double>::max(),
                                    std::numeric_limits<double>::max(),
                                    std::numeric_limits<double>::max()}};
  auto high = std::array<double, 3>{{std::numeric_limits<double>::lowest(),
                                     std::numeric_limits<double>::lowest(),
                                     std::numeric_limits<double>::lowest()}};
  for (const atom& a : atoms) {
    const auto& x = a.position;
    if (x[0] < low[0])
      low[0] = x[0];
    if (x[1] < low[1])
      low[1] = x[1];
    if (x[2] < low[2])
      low[2] = x[2];
    if (x[0] > high[0])
      high[0] = x[0];
    if (x[1] > high[1])
      high[1] = x[1];
    if (x[2] > high[2])
      high[2] = x[2];
  }
  auto box_size = geom::sub(high, low);
  // ensure a minimum size
  geom::clamp_min(box_size, 10.0);
  geom::scale(box_size, box_scale, box_size);

  // create p1 cell
  cell =
      new unit_cell(1, box_size[0], box_size[1], box_size[2], 90.0, 90.0, 90.0);
  auto hmax = 0;
  auto kmax = 0;
  auto lmax = 0;
  auto res_low = std::numeric_limits<double>::max();
  auto res_high = res_low;
  for (hmax = 1; cell->resolution(hmax, 0, 0) >= resolution_limit; hmax++)
    ;
  auto res = cell->resolution(hmax, 0, 0);
  if (res < res_high)
    res_high = res;

  for (kmax = 1; cell->resolution(0, kmax, 0) >= resolution_limit; kmax++)
    ;
  res = cell->resolution(0, kmax, 0);
  if (res < res_high)
    res_high = res;

  for (lmax = 1; cell->resolution(0, 0, lmax) >= resolution_limit; lmax++)
    ;
  res = cell->resolution(0, 0, lmax);
  if (res < res_high)
    res_high = res;

  crystallography::fourier_map<float> map(*cell, res_low, res_high,
                                          fft_sampling);
  for (const auto& a : atoms) {
    // fractionalize the model coordinates relative to the p1 cell
    std::array<double, 3> frac = cell->fractionalize(a.position);
    map.add_atom(frac[0], frac[1], frac[2], a.occ, a.b, a.Z);
  }
  map.invert();

  sfgrid = new grid3<std::complex<T>>(
      static_cast<float>(-hmax), static_cast<float>(-kmax), -1.0f, 1.0f, 1.0f,
      1.0f, 2 * hmax + 1, 2 * kmax + 1, lmax + 2);
  for (auto ih = -hmax; ih <= hmax; ih++) {
    for (auto ik = -kmax; ik <= kmax; ik++) {
      for (auto il = -1; il <= lmax; il++) {
        if (ih == 0 && ik == 0 && il == 0)
          continue;
        auto res = cell->resolution(ih, ik, il);
        // check that the reflection is within the bounds of the map
        if (res > res_high) {
          std::complex<T> f{map.get_structure_factor(ih, ik, il, res)};
          sfgrid->set_value(ih, ik, il, f);
        }
      }
    }
  }
}

template <class T>
molecular_transform<T>::~molecular_transform() {
  delete sfgrid;
  delete cell;
}
}
#endif
